import os
import numpy as np
import pandas as pd
import time
import json
import pathlib
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedShuffleSplit
from numpy import loadtxt, savetxt
from sklearn import metrics, model_selection
    
from cnn.train_utils import *
from cnn.models import *
from cnn.metrics import *

curr_dir_path = str(pathlib.Path().absolute())
cwd = os.getcwd() 
save_dir_path = '/cnn/' + '/logs/'
gpu_num = 0 
cuda_available = torch.cuda.is_available()
if cuda_available:
    torch.cuda.set_device(gpu_num)
    
class Training():

    def __init__(self, config, dim, save_dir = '', device='', model_no=0):
        self.config = config
        
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.save_dir = save_dir
        self.model_no = model_no

        # self.model = CNN(self.device)
    
        #hyperparameter tuning class test 
        """
        self.model = CNN_hp_tuning(self.device, no_conv=config['MODEL']['Conv no'], no_filters=config['MODEL']['Filter no'], 
                                   filter_size=config['MODEL']['Filter size'], stride=config['MODEL']['Conv stride'], 
                                   pool_size=config['MODEL']['Pool size'], pool_stride=config['MODEL']['Pool stride'],
                                   dense_output=config['MODEL']['FC layer output'], no_dense_layers = config['MODEL']['No FC layers'],
                                   dropout=config['MODEL']['dropout'])
        """
        self.model = MLP(input_dim=dim)
        
        if cuda_available:
            print("Model on GPU")
            self.model.cuda()
  
        self.optimizer = getattr(optim, config['OPTIMIZER']['type']) \
            (self.model.parameters(), lr=config['OPTIMIZER']['lr'], weight_decay=config['OPTIMIZER']['weight_decay'])

        
        self.trainloader = None
        self.metrics = {'train': {}, 'val': {}}
        self.train_accuracy_curve = []
        self.train_f1_curve = []
        self.val_accuracy_curve  = []
        self.val_f1_curve = []

    def train(self, epoch, x_train, y_train):
        
        trainset = SequenceDataset(x_train, y_train)  
        
        self.trainloader = torch.utils.data.DataLoader(
                        trainset, batch_size=self.config['DATA']['BATCH_SIZE'], shuffle=True)
                     #   shuffle=self.config['DATA']['SHUFFLE'])#, num_workers=self.config['DATA']['NUM_WORKERS'])

        self.model.to(self.device)
        self.model.train()
        
        loss_function = torch.nn.BCELoss()

        # Metrics
        m = Metrics()  
        self.metrics['train'] = m.metrics      
        avg_train_loss = 0
    
        for data, labels  in self.trainloader:     
            
            self.optimizer.zero_grad()
            
            output = self.model.forward(data.to(self.device))     
          #  output = output.squeeze(1)
            output = torch.squeeze(output, 1)
            
            labels = torch.squeeze(labels, 1)

            loss = loss_function(output, labels.to(self.device))
              
            #Backpropagation + optimization step      
            loss.backward()
            self.optimizer.step()
            
            #Metrics per batch
            metrics_for_batch, pred = m.get_metrics(output.detach().clone(), labels.detach().clone())  
            for key,value in metrics_for_batch.items():
                self.metrics['train'][key] += value
                
            avg_train_loss += loss.item()
            
            
        # Metrics
        for measure in m.metrics:
            self.metrics['train'][measure] /= (len(self.trainloader)+1)
        
        avg_train_loss /= (len(self.trainloader))
        self.metrics['train']['loss'] = avg_train_loss        
        print('Epoch: {:d}, Train Loss: {:.4f}, '.format(epoch, avg_train_loss), self.metrics['train'])
        
        return self.metrics['train']

    def evaluate(self, epoch, x_val, y_val, testing=False):
        valset = SequenceDataset(x_val, y_val)  
        val_dataloader = torch.utils.data.DataLoader(
                    valset, batch_size=self.config['DATA']['BATCH_SIZE'], shuffle=True)
                  #  shuffle=self.config['DATA']['SHUFFLE'])#, num_workers=self.config['DATA']['NUM_WORKERS'])
        
        self.model.to(self.device)
        self.model.eval()
        
        loss_function = torch.nn.BCELoss()
        
        # Metrics
        avg_val_loss = 0
        
        m = Metrics()  
        self.metrics['val'] = m.metrics
                
        for data, labels  in val_dataloader:                 
            
            output = self.model.forward(data.to(self.device))           
            output = torch.squeeze(output, 1)
            labels = torch.squeeze(labels, 1)
            
            loss = loss_function(output, labels.to(self.device))
            
            #Metrics per batch
            metrics_for_batch, pred = m.get_metrics(output.detach().clone(), labels.detach().clone())  
            for key,value in metrics_for_batch.items():
                self.metrics['val'][key] += value
                
            avg_val_loss += loss.item()
            
        #Metrics
        for measure in m.metrics:
            self.metrics['val'][measure] /= (len(val_dataloader)+1)
        
        avg_val_loss /= (len(val_dataloader))
        self.metrics['val']['loss'] = avg_val_loss
        
        if testing == False:
            print('Epoch: {:d}, Valid Loss: {:.4f}, '.format(epoch, avg_val_loss), self.metrics['val'])
        else:
            print('Epoch: {:d}, Test Loss: {:.4f}, '.format(epoch, avg_val_loss), self.metrics['val'])

        return self.metrics['val']


    
    def save_data(self):
        # Save metrics and model
        if self.config['TRAINER']["save_model"]:
            json_dir = cwd + self.save_dir + f"model{self.model_no}.pt"
            print("Saving data at: " + json_dir)
            torch.save(self.model.state_dict(), json_dir)
            #to load model:
            #json_dir = self.save_dir + "model.pt"
            #model = CNN(args)
            #model.load_state_dict(torch.load(json_dir))
            
            #Save final metrics 
            results = [self.metrics, self.config]
            json_dir = cwd + self.save_dir + "final_metrics"
            print("Saving data at: " + json_dir)
            output_folder = pathlib.Path(json_dir)
            output_folder.mkdir(parents=True, exist_ok=True)
            results_json = json.dumps(results, indent=4, sort_keys=True)
            (output_folder / f"results{self.model_no}.json").write_text(results_json)
            
            train_accuracy_curve = np.hstack(self.train_accuracy_curve)
            train_f1_curve = np.hstack(self.train_f1_curve)
            val_accuracy_curve = np.hstack(self.val_accuracy_curve)
            val_f1_curve = np.hstack(self.val_f1_curve)
            #Save training and validation curves (accuracy, f1)
            results = {"train_accuracy": train_accuracy_curve.tolist(), "train_f1":  train_f1_curve.tolist(), 
                        "val_accuracy": val_accuracy_curve.tolist(), "val_f1":  val_f1_curve.tolist()}            
            json_dir_curve = cwd + self.save_dir + "curves"
            print("Saving training and validation curves at: " + json_dir_curve)
            output_folder = pathlib.Path(json_dir_curve)
            output_folder.mkdir(parents=True, exist_ok=True)
            results_json = json.dumps(results, indent=4, sort_keys=True)
            (output_folder / f"results{self.model_no}.json").write_text(results_json)
            
            
    def training_pipeline(self,  train_data, train_labels, val_data, val_labels, test_data, test_labels):
        
        train_accuracy_curve = []
        train_f1_curve = []
        val_accuracy_curve = []
        val_f1_curve = []
        
           
        for epoch in range(self.config['TRAINER']['epochs']):
            print("Training Epoch %i -------------------" % epoch)

            epoch_tic = time.time()
            metrics = self.train(epoch, train_data, train_labels)
            self.train_accuracy_curve.append(metrics['acc'])
            self.train_f1_curve.append(metrics['f1'])
            
            if self.config['VALIDATION']['apply']:
                metrics = self.evaluate(epoch, val_data, val_labels)
                self.val_accuracy_curve.append(metrics['acc'])
                self.val_f1_curve.append(metrics['f1'])
                
            epoch_toc = time.time()
            epoch_time = epoch_toc - epoch_tic

            print("--------------Epoch %i completed in %i seconds--------------" % (epoch, epoch_time))
          
        #Testing
        metrics = self.evaluate(epoch, test_data, test_labels, testing = True)
        self.save_data()
        
        return self.train_accuracy_curve, self.train_f1_curve, self.val_accuracy_curve, self.val_f1_curve
    

       